#include <math.h>
#include "process.h"
#include "test.h"
#include "../../../../process.h"
#include "../../../../system.h"
#include "../../../../unittest.h"
#include "../../../../interfaces/network_rx_interface/process.h"
#include "../../../../interfaces/network_tx_interface/process.h"
#include "../../../../databases/ink_paper_state_database/process.h"
#include "../../../../databases/job_queue/process.h"
#include "../../../../databases/user_database/process.h"

static int _get_empty_job_index();
static int _get_require_ink(char *, int);
static int _get_require_paper(char *, int);
static int _get_job_queue_require_ink();
static int _get_job_queue_require_paper();
static void trigger(struct Packet *);
static int last_job_id;

static struct PrintValidationProcess print_validation_process = {
    /* Private Functions : Not accesable to other processes */
    ._get_empty_job_index = _get_empty_job_index,
    ._get_require_ink = _get_require_ink,
    ._get_require_paper = _get_require_paper,
    ._get_job_queue_require_ink = _get_job_queue_require_ink,
    ._get_job_queue_require_paper = _get_job_queue_require_paper,
    /* Public Functions : Accessable to other processes */
    .trigger = trigger
};

void init_print_validation_process() {
    printf("Initialization Validation Process...\t\t");
    set_process(PRINT_VALIDATION_PROCESS, &print_validation_process);
    SET_UNIT_TEST(PRINT_VALIDATION_PROCESS, print_validation_process_unit_test);
    printf("\033[92m[OK]\033[0m\n");
    last_job_id = 0;
}

static int _get_empty_job_index() {
    int i;
    struct JobQueue *job_queue = get_process(JOB_QUEUE);
    
    for(i=0; i<JOB_MAX_SIZE; i++) if(job_queue->jobs_info[i].job_id == -1) return i;
    return -1;
}

static int _get_require_ink(char *buf, int size) {
    int i;
    int require_ink = 0;
    int line_text_length = 0;
    int sp_count = 0;
    
    for(i=0; i<size; i++) {
        if(buf[i] != '\n') {
            if(line_text_length < 30) {
                line_text_length += 1;
                if(buf[i] == ' ') sp_count += 1;
            }
        } else {
            require_ink += line_text_length - sp_count;
            line_text_length = 0;
            sp_count = 0;
        }
    }
    
    return require_ink;
}

static int _get_require_paper(char *buf, int size) {
    int i;
    int line_count = 0;
    
    for(i=0; i<size; i++) {
        if(buf[i] == '\n') line_count += 1;
    }
    
    return (int)ceil(line_count / 10.0);
}

static int _get_job_queue_require_ink() {
    int require_ink = 0;
    int i;
    struct JobQueue *job_queue = get_process(JOB_QUEUE);
    
    for(i=0; i<JOB_MAX_SIZE; i++) if(job_queue->jobs_info[i].job_id != -1) {
        require_ink += _get_require_ink(job_queue->jobs_info[i].buffer, job_queue->jobs_info[i].file_size);
    }
    
    return require_ink;
}

static int _get_job_queue_require_paper() {
    int require_paper = 0;
    int i;
    struct JobQueue *job_queue = get_process(JOB_QUEUE);
    
    for(i=0; i<JOB_MAX_SIZE; i++) if(job_queue->jobs_info[i].job_id != -1) {
        require_paper += job_queue->jobs_info[i].total_page;
    }
    
    return require_paper;
}

static void trigger(struct Packet *packet){
    const int start = TYPE_LENGTH + REQ_ID_LENGTH;
    struct InkPaperStateDatabase *ink_paper_db = get_process(INK_PAPER_STATE_DATABASE);
    struct JobQueue *job_queue = get_process(JOB_QUEUE);
    int index;
    struct Packet res;
    char res_data[TYPE_LENGTH + REQ_ID_LENGTH + RES_TYPE_LENGTH + JOB_ID_LENGTH];
    memcpy(res_data, packet->data, TYPE_LENGTH + REQ_ID_LENGTH);
    
    res.data = res_data;
    res.length = TYPE_LENGTH + REQ_ID_LENGTH + RES_TYPE_LENGTH + JOB_ID_LENGTH;
    res.sock = packet->sock;
    
    index = _get_empty_job_index();
    
    if(index == -1) { // 큐가 꽉찬 경우
        // res_type
        res_data[start] = 1;
        // job_id
        res_data[start+1] = 0;
        res_data[start+2] = 0;
        res_data[start+3] = 0;
        res_data[start+4] = 0;
        
        ((struct NetworkTxInterface *)get_process(NETWORK_TX_INTERFACE))->send_packet(res);
        return;
    }
    
    if(packet->length < TYPE_LENGTH + REQ_ID_LENGTH + USER_ID_MAX_LENGTH + USER_PW_MAX_LENGTH + 4) {
        printf("Print Length Error 1 - %d", packet->length);
        return;
    }
    
    char *file_size_buf = packet->data + TYPE_LENGTH + REQ_ID_LENGTH + USER_ID_MAX_LENGTH + USER_PW_MAX_LENGTH;
    int file_size = ((file_size_buf[0] << 24)&0xff000000) | ((file_size_buf[1] << 16)&0xff0000) | ((file_size_buf[2] << 8)&0xff00) | (file_size_buf[3]&0xff);
    char *file_packet_buf = packet->data + TYPE_LENGTH + REQ_ID_LENGTH + USER_ID_MAX_LENGTH + USER_PW_MAX_LENGTH + 4;

    printf("file_size:%d\n", file_size);
    if(packet->length < TYPE_LENGTH + REQ_ID_LENGTH + USER_ID_MAX_LENGTH + USER_PW_MAX_LENGTH + 4 + file_size) {
        printf("Print Length Error 2 - %d", packet->length);
        return;
    }
    
    char *file_buf = (char *)malloc(sizeof(char) * (file_size+1));
    memcpy(file_buf, packet->data + TYPE_LENGTH + REQ_ID_LENGTH + USER_ID_MAX_LENGTH + USER_PW_MAX_LENGTH + 4, file_size);
    file_buf[file_size] = '\n';
    file_size += 1;
    
    // 잉크가 부족한 경우
    if((_get_job_queue_require_ink() + _get_require_ink(file_buf, file_size)) > *(ink_paper_db->current_ink_value)) {
        free(file_buf);
        // res_type
        res_data[start] = 2;
        // job_id
        res_data[start+1] = 0;
        res_data[start+2] = 0;
        res_data[start+3] = 0;
        res_data[start+4] = 0;
        
        ((struct NetworkTxInterface *)get_process(NETWORK_TX_INTERFACE))->send_packet(res);
        
        return;
    }
    
    // 용지가 부족한 경우
    if((_get_job_queue_require_paper() + _get_require_paper(file_buf, file_size)) > *(ink_paper_db->current_paper_value)) {
        free(file_buf);
        // res_type
        res_data[start] = 3;
        // job_id
        res_data[start+1] = 0;
        res_data[start+2] = 0;
        res_data[start+3] = 0;
        res_data[start+4] = 0;
        
        ((struct NetworkTxInterface *)get_process(NETWORK_TX_INTERFACE))->send_packet(res);
        
        return;
    }
    
    // create job
    
    job_queue->jobs_info[index].job_id = ++last_job_id;
    job_queue->jobs_info[index].state = 0;
    memcpy(job_queue->jobs_info[index].user_id, packet->data + TYPE_LENGTH + REQ_ID_LENGTH, 12);
    job_queue->jobs_info[index].current_page = 0;
    job_queue->jobs_info[index].total_page = _get_require_paper(file_packet_buf, file_size);
    job_queue->jobs_info[index].file_size = file_size;
    job_queue->jobs_info[index].buffer = file_buf;
    job_queue->jobs_info[index].fd = -1;
    
     // res_type
    res_data[start] = 0;
    // job_id
    res_data[start+1] = (job_queue->jobs_info[index].job_id >> 24) & 0xff;
    res_data[start+2] = (job_queue->jobs_info[index].job_id >> 16) & 0xff;
    res_data[start+3] = (job_queue->jobs_info[index].job_id >> 8) & 0xff;
    res_data[start+4] = job_queue->jobs_info[index].job_id & 0xff;
    
    ((struct NetworkTxInterface *)get_process(NETWORK_TX_INTERFACE))->send_packet(res);
}
